/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */


package org.apache.geronimo.jetty6.handler;

import java.io.IOException;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.transaction.Status;
import javax.transaction.SystemException;
import javax.transaction.UserTransaction;

import org.mortbay.jetty.handler.AbstractHandler;

/**
 * @version $Rev: 535017 $ $Date: 2007-05-03 19:10:58 -0400 (Thu, 03 May 2007) $
 */
public class UserTransactionHandler extends AbstractImmutableHandler {
    private final UserTransaction userTransaction;

    public UserTransactionHandler(AbstractHandler next, UserTransaction userTransaction) {
        super(next);
        this.userTransaction = userTransaction;
    }

    public void handle(String target, HttpServletRequest request, HttpServletResponse response, int dispatch) throws IOException, ServletException {
        boolean active = isActive();
        try {
            next.handle(target, request, response, dispatch);
        } finally {
             if ((!active && isMarkedRollback()) || (dispatch == REQUEST && isActive())) {
                try {
                    userTransaction.rollback();
                } catch (SystemException e) {
                    throw new ServletException("Error rolling back transaction left open by user program", e);
                }
            }
        }
    }

    public void lifecycleCommand(LifecycleCommand lifecycleCommand) throws Exception {
        boolean active = isActive();
        try {
            super.lifecycleCommand(lifecycleCommand);
        } finally {
            if (!active && isActive()) {
                try {
                    userTransaction.rollback();
                } catch (SystemException e) {
                    throw new ServletException("Error rolling back transaction left open by user program", e);
                }
            }
        }
    }

    private boolean isActive() throws ServletException {
        try {
            return !(userTransaction.getStatus() == Status.STATUS_NO_TRANSACTION
                    || userTransaction.getStatus() == Status.STATUS_COMMITTED);
        } catch (SystemException e) {
            throw new ServletException("Could not determine transaction status", e);
        }
    }
    private boolean isMarkedRollback() throws ServletException {
        try {
            return userTransaction.getStatus() == Status.STATUS_MARKED_ROLLBACK;
        } catch (SystemException e) {
            throw new ServletException("Could not determine transaction status", e);
        }
    }
}
